import ij.*;
import ij.process.*;
import ij.gui.*;
import ij.io.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.imageio.ImageIO;
import javax.swing.*;
import java.util.*;
import ij.plugin.*;
import ij.plugin.frame.*;
import ij.text.*;
import ij.measure.*;

/** The plugin extracts plotable data and using a GnuPlot script 
 *  and runs them through gnuplot, displaying the results in ImageJ.
 *
 * @author Fred Damen <fred@damen.org>
 *
 * Version History
 * 2020-04-01 Released
 */
public class Gnu_Plot implements PlugIn, DialogListener {
   Editor neweditor = null;
   Vector<Editor>      ev = new Vector<Editor>();
   Vector<ImagePlus>   iv = new Vector<ImagePlus>();

   Panel                          dspw = new Panel();
   Vector<Choice>                 dcwv = new Vector<Choice>();
   HashMap<Choice,Vector<Frame>> dscfm = new HashMap<Choice,Vector<Frame>>();

   /** gui */
   GenericDialog gd = null;
   public void run(String arg) {

      if (arg.equals("options")) {
         gnuplot(null);
         return;
         }

      gd = new NonBlockingGenericDialog("GnuPlot");

      gd.addStringField("Plot Title", "Gnuplot");
      ((TextField)gd.getStringFields().get(0)).setColumns(15);
      gd.addChoice("GNUPlot Script:",new String[]{"<choose/rescan>"},"");
      Choice gpscw = (Choice)gd.getChoices().get(0);
      populateScriptsChoice(gpscw);
      gpscw.addItemListener(new ItemListener() { 
         @Override public void itemStateChanged(ItemEvent e) {
            if (((Choice)e.getSource()).getSelectedIndex() == 0) {
               populateScriptsChoice((Choice)e.getSource());
               neweditor = null;
               gd.pack();
               }
            }
         });

      gd.addPanel(dspw,GridBagConstraints.NORTH, new Insets(0, 0, 0, 0));
      dspw.setLayout(new GridBagLayout());
      addNewData();

      gd.addMessage("Things to do . . . ");
      gd.addCheckbox("Generate Plot",false);       gd.addToSameRow(); gd.addCheckbox("Erase Extra Images",false);
      gd.addCheckbox("View Complete Input",false); gd.addToSameRow(); gd.addStringField("#", "x,y,z");
      ((TextField)gd.getStringFields().get(1)).setColumns(15);
      gd.addCheckbox("Test commpatability",false); gd.addToSameRow(); gd.addStringField("Terminal", "png");
      ((TextField)gd.getStringFields().get(2)).setColumns(15);

      gd.addMessage("Hmm...\n     Check the Log window.\nOr http://www.gnuplot.info/");

      gd.getButtons()[0].setEnabled(false);

      gd.addDialogListener(this);
      gd.showDialog();
      if (gd.wasCanceled()) return;
      }

   private void populateScriptsChoice(Choice c) {
      int csi = c.getSelectedIndex();
      Editor current = csi > 2 ? ev.get(csi-3) : null;

      c.removeAll();
      c.add("<choose/rescan>");
      c.add("<new>");
      c.add("<new ask>");
      ev.clear();
      Frame[] fs = WindowManager.getNonImageWindows();
      if (fs == null || (fs.length == 1))
         return;
      for(Frame f : fs) 
              if (f.getTitle().endsWith(".java"))   continue;
         else if (f.getTitle().equals("Exception")) continue;
         else if (f.getTitle().equals("Errors"))    continue;
         else if (f.getClass().getName().endsWith("Editor")) {
            if (f == current) csi = c.getItemCount();
            c.add(f.getTitle());
            ev.add((Editor)f);
            }
      c.select(csi);
      }
   private void populateDataChoice(Choice c) {
      int csi = Math.max(0,c.getSelectedIndex());
      Frame current = csi > 2 ? dscfm.get(c).get(csi-3) : null;

      c.removeAll();
      c.add("<none/rescan>");
      c.add("<another>");
      dscfm.put(c, new Vector<Frame>());
      Frame[] fs = WindowManager.getNonImageWindows();
      if (fs == null || (fs.length == 1))
         return;
      for(Frame f : fs)
              if (f.getTitle().endsWith(".java"))   continue;
         else if (f.getTitle().equals("Exception")) continue;
         else if (f.getTitle().equals("Errors"))    continue;
         else if (f.getClass().getName().endsWith("Editor")) continue;
         else if (f.getClass().getName().endsWith("TextWindow")) {
            if (f == current) csi = c.getItemCount();
            c.add(f.getTitle());
            dscfm.get(c).add((Frame)f);
            }
      int[] ids = WindowManager.getIDList();
      if (ids == null) 
         return;
      for(int id : ids) {
         ImageWindow iw = WindowManager.getImage(id).getWindow();
         if (iw.getClass().getName().endsWith("PlotWindow")) {
            if (iw == current) csi = c.getItemCount();
            c.add(iw.getTitle());
            dscfm.get(c).add((Frame)iw);
            }
         }
      c.select(csi);
      c.addItemListener(new ItemListener() { 
         @Override public void itemStateChanged(ItemEvent e) {
            int si = ((Choice)e.getSource()).getSelectedIndex();
            if (si == 0) populateDataChoice((Choice)e.getSource());
            if (si == 1) addNewData();
            gd.pack();
            }
         });
      
      }
   private void addNewData() {
      GridBagConstraints gbc = new GridBagConstraints();
      gbc.fill = GridBagConstraints.VERTICAL;
      gbc.weightx = 0.5;

      gbc.gridy = dcwv.size();
      gbc.gridx = 0;
      dspw.add(new Label("          $data"+dcwv.size(),Label.RIGHT),gbc);
      gbc.gridx = 1;
      Choice c = new Choice();
      populateDataChoice(c);
      c.addItemListener(new ItemListener() { 
         @Override public void itemStateChanged(ItemEvent e) {
            if (((Choice)e.getSource()).getSelectedIndex() == 0) {
               populateDataChoice((Choice)e.getSource());
               gd.pack();
               }
            }
         });
      dspw.add(c, gbc);
      dcwv.add(c);
      }

   public boolean dialogItemChanged(GenericDialog gd, AWTEvent aev) {
      if (aev==null) return false;

      Point loc = gd.getLocation();
      Dimension ws = gd.getSize();
      ImageWindow.setNextLocation(loc.x+ws.width, loc.y+0*ws.height);

      String title = gd.getNextString(); 
      String code = null;
      int ci = gd.getNextChoiceIndex();
      if (ci == 0) ; //return false;
      else if (ci <= 2) {
         if (neweditor == null || !neweditor.isShowing()) {
            neweditor = new Editor(20,80,22,Editor.MONOSPACED+Editor.MENU_BAR);
            Dimension news = neweditor.getSize();
            neweditor.setLocation(Math.max(0,loc.x-news.width),loc.y);
            neweditor.display(title+".gp","set terminal png\nplot \"$data0\"");
            if (ci == 2) {
               IJ.wait(1000);
               OpenDialog od = new OpenDialog("GnuPlot Scipt File");
               if (od.getFileName() != null)
                  neweditor.open(od.getDirectory(),od.getFileName());
               }
            else return false;
            }
         else {
            code  = neweditor.getText();
            title = neweditor.getTitle().replaceAll(".gp$","");
            }
         }
      else {
         code = ev.get(ci-3).getText();
         title = ev.get(ci-3).getTitle().replaceAll(".gp$","");
         }

      Vector<String> dv = new Vector<String>();
      for(Choice dc : dcwv) {
         int di = dc.getSelectedIndex()-2;
         if (di < 0) continue;
         Frame f = dscfm.get(dc).get(di);
              if (f.getClass().getName().endsWith("TextWindow")) dv.add(((TextWindow)f).getTextPanel().getText());
         else if (f.getClass().getName().endsWith("PlotWindow")) {
            ResultsTable rt = ((PlotWindow)f).getResultsTable();
            StringBuffer sb = new StringBuffer();
            for(int i=0; i<rt.size(); i++)
               sb.append(rt.getRowAsString(i)+"\n");
            dv.add(sb.toString());
            }
         }

      String gps = joinCodeAndData(code,dv.toArray(new String[0]));

      Checkbox preview = (Checkbox)gd.getCheckboxes().get(0);
      if (preview.getState()) {
         preview.setState(false);
         ImagePlus imp = process(gps,title);
         if (imp == null) return false;
         imp.show();
         iv.add(imp);
         }

      Checkbox edwin = (Checkbox)gd.getCheckboxes().get(2);
      String flc = "# "+gd.getNextString();
      if (edwin.getState()) {
         edwin.setState(false);
         if (gps == null)
            return true;
         Editor e = (new Editor(20,80,22,Editor.MONOSPACED+Editor.MENU_BAR));
         e.display(title+"_wd.gp",flc+"\n"+gps);
         e.setLocation(loc.x+ws.width, loc.y+0*ws.height);
         return true;
         }

      Checkbox testwin = (Checkbox)gd.getCheckboxes().get(3);
      String term = "set terminal "+gd.getNextString();
      if (testwin.getState()) {
         testwin.setState(false);
         return processAndDisplay(term+"\ntest\n","Test");
         }

      Checkbox erase = (Checkbox)gd.getCheckboxes().get(1);
      for(int i=iv.size()-2; i>= 0; i--)
              if (iv.get(i) == null || !iv.get(i).isVisible()) iv.remove(i);
         else if (erase.getState())                            iv.remove(i).close();

      return false;
      }

   /** format code and data suitable for gnuplot processing of stdin.
    */
   public static String joinCodeAndData(String code, String[] data) {
      if (code == null)
         return null;

      if (data.length < 1)
         return code;

      String gps = "";
      for(int i=0; i<data.length; i++)
         gps += "$data"+i+" << EOD\n"+data[i].replaceAll("\\t\\t","\tNaN\t").replaceAll("\\t\\t","\tNaN\t")+"EOD\n";
      gps += code;
      return gps;
      }

   private static String gnuplot = gnuplot("/usr/bin/gnuplot");   // if all else fails . . .
   /** configured */
   public static String gnuplot() { return new String(gnuplot); }
   /** configure, null - ask, "" ask if current bad, otherwise set/save if good. */
   public static String gnuplot(String gp) {
      String gpp = Prefs.get("gnuplot.cmd","");
           if (gp == null)                        gpp = null;
      else if (gp.equals(""))                     gpp = (gnuplot != null) && (new File(gnuplot)).exists() ? gnuplot : null;
      else if (gnuplot == null && gpp != null)    gpp = gpp;
      else                                        gpp = gp;

      if (gpp != null && !(new File(gpp).exists()))
         gpp = null;

      if (gpp == null)
         gpp = (new OpenDialog("Select GnuPlot executable","")).getPath();

      if (gpp == null)
         return null;

      if (!(new File(gpp)).exists()) {
         IJ.error("Gnu_Plot: The requested executable '"+gpp+"' does not exist.");
         return null;
         }

      gnuplot = gpp;
      
      Prefs.set("gnuplot.cmd",gpp);
      Prefs.savePreferences();
      return gpp;
      }

   /** run gps script through gnuplot and get image back
    */
   public static ImagePlus process(String gps, String title) {
      if (gps == null) {
         IJ.error("Gnu_Plot: Need script");
         return null;
         }                   

      try {
         Process process = new ProcessBuilder(gnuplot,"-p").start();
         OutputStream os = process.getOutputStream();
         os.write(gps.getBytes());
         os.close();
         InputStream is = process.getInputStream();
         ImagePlus gimp = new ImagePlus(title, ImageIO.read(is));
         BufferedReader ebr = new BufferedReader(new InputStreamReader(process.getErrorStream()));
         String str = null;
         while((str = ebr.readLine()) != null)
            IJ.log(str);
         return gimp;
         }
      catch(IOException ioe) {
         StringWriter sw = new StringWriter();
         ioe.printStackTrace(new PrintWriter(sw));
         IJ.error(sw.toString());  
         return null;
         }
      }

   /** and display.
    */
   public static boolean processAndDisplay(String gps, String title) {
      ImagePlus imp = process(gps,title);
      if (imp == null) return false;
      imp.show();
      return true;
      }

   /** run gps script through gnuplot and leave running, close OutputStream to close gnuplot process.
    */
//volatile OutputStream os = null;
   public  void processing(final String gps, String title) {
      if (gps == null) {
         IJ.error("Gnu_Plot: Need script");
         return ;//null;
         }


//(new Thread() { public void run() {
      try {
         ProcessBuilder pb = new ProcessBuilder(gnuplot,"-p"); 
         for(Map.Entry<String,String> e : pb.environment().entrySet())
            IJ.log(e.getKey()+" = "+e.getValue());
         Process process = pb.start();
         OutputStream os = process.getOutputStream();
         os.write(gps.getBytes());
         //os.close();
         //InputStream is = process.getInputStream();
         //ImagePlus gimp = new ImagePlus(title, ImageIO.read(is));
         final BufferedReader ibr = new BufferedReader(new InputStreamReader(process.getInputStream()));
//         (new Thread() { 
//            public void run() {
               try {
                  String str = null;
                  while((str = ibr.readLine()) != null)
                     IJ.log(str);
                  }
               catch(IOException ioe) {
                  StringWriter sw = new StringWriter();
                  ioe.printStackTrace(new PrintWriter(sw));
	          IJ.log(sw.toString());  
                  }
//               }
//            }).start();

         final BufferedReader ebr = new BufferedReader(new InputStreamReader(process.getErrorStream()));
//         (new Thread() { 
//            public void run() {
               try {
                  String str = null;
                  while((str = ebr.readLine()) != null)
                     IJ.log(str);
                  }
               catch(IOException ioe) {
                  StringWriter sw = new StringWriter();
                  ioe.printStackTrace(new PrintWriter(sw));
	          IJ.log(sw.toString());  
                  }
//               }
//            }).start();
         //return os;
         }
      catch(IOException ioe) {
         IJ.error("IOE"+ioe);
         //return null;
         }
//               }
//            }).start();
      IJ.wait(100);
//      return os;
      }

   /** utility to convert data for input to gnuplot */
   public static <T> String arraysToString(int n, T[][] array) {
      String data = "$data"+n+" <<EOD\n";
      for(int i=0; i<array[0].length; i++, data+="\n")
         for(int j=0; j<array.length; j++)
            data += array[j][i]+"\t";
      data += "EOD\n\n";
      return data;
      }
   /** produce input to gnuplot. */
   public static <T> String plotgs(String code, T[][]... arrs) {
      String data = "";
      int i = 0;
      for(T[][] arr : arrs)
         data += arraysToString(i++, arr);
      return data+code;
      }
   /** code/data in image out. */
   public static <T> ImagePlus plot(String title, String code, T[][]... arrs) {
      return process(plotgs(code,arrs),title);
      }

   /** utility to convert data for input to gnuplot */
   public static String arraysToString(int n, double[][] arr) {
      String data = "$data"+n+" <<EOD\n";
      for(int i=0; i<arr[0].length; i++, data+="\n")
         for(int j=0; j<arr.length; j++)
            data += arr[j][i]+"\t";
      data += "EOD\n\n";
      return data;
      }
   /** produce input to gnuplot. */
   public static String plotgs(String code, double[][]... arrs) {
      String data = "";
      int i = 0;
      for(double[][] arr : arrs)
         data += arraysToString(i++, arr);
      return data+code;
      }
   public static ImagePlus plot(String title, String code, double[][]... arrs) {
      return process(plotgs(code,arrs),title);
      }

   /** utility to convert data for input to gnuplot */
   public static String arraysToString(int n, float[][] arr) {
      String data = "$data"+n+" <<EOD\n";
      for(int i=0; i<arr[0].length; i++, data+="\n")
         for(int j=0; j<arr.length; j++)
            data += arr[j][i]+"\t";
      data += "EOD\n\n";
      return data;
      }
   /** produce input to gnuplot. */
   public static String plotgs(String code, float[][]... arrs) {
      String data = "";
      int i = 0;
      for(float[][] arr : arrs)
         data += arraysToString(i++, arr);
      return data+code;
      }
   /** code/data in image out. */
   public static ImagePlus plot(String title, String code, float[][]... arrs) {
      return process(plotgs(code,arrs),title);
      }

   /** code/data in gnuplot gui left running, close OutputStream to close fnuplot gui. */
   public static void plotting(String title, String code, float[][]... arrs) {
      (new Gnu_Plot()).processing(plotgs(code,arrs),title);
      return;
      }
}
